;;;; (net http message) -- an interface to the HTTP protocol. 

;;  Copyright 2001, Evan Prodromou <evan@prodromou.san-francisco.ca.us>.

;;  This program is free software; you can redistribute it and/or modify
;;  it under the terms of the GNU General Public License as published by
;;  the Free Software Foundation; either version 2 of the License, or
;;  (at your option) any later version.

;;  This program is distributed in the hope that it will be useful,
;;  but WITHOUT ANY WARRANTY; without even the implied warranty of
;;  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;  GNU General Public License for more details.

;;  You should have received a copy of the GNU General Public License
;;  along with this program; if not, write to the Free Software
;;  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

(define-module (net http message)
  :use-module (net http util)
  :use-module (oop goops)
  :use-module (oop goops describe)
  :use-module (ice-9 format)
  :use-module (ice-9 regex))

;; Compatibility kludge.

(or (defined? 'read-line)
    (use-modules (ice-9 rdelim)))

(export	<http-request>
	<http-response>
	header
        header-set!
        has-header
	version
	headers
	content
	meth
	uri
	server
	port
	code
	status
	send-request
	to-port
	request-from-port
	response-from-port)

;;; The version of this module

(define *module-version* "0.3.1")

;;; Class defining an HTTP message, superclass of requests and
;;; responses.

(define-class <http-message> (<object>)
  (version #:init-value "1.0"		; string
           #:accessor version
	   #:init-keyword #:version)
  (headers #:init-value '()		; assoc list
           #:accessor headers
	   #:init-keyword #:headers)
  (content #:init-value '()		; port
	   #:accessor content
	   #:init-keyword #:content))

;;; Convenience method for setting a message header.

(define-method (header-set! (msg <http-message>) name value)
  (set! (headers msg) (assoc-set! (headers msg) name value)))

;;; Convenience method for getting a message header.

(define-method (header (msg <http-message>) name)
  (let ((pair (assoc name (headers msg))))
    (if pair
	(cdr pair)
	#f)))

;;; Convenience method to check the existence of a header.

(define-method (has-header (msg <http-message>) name)
  (if (assoc name (headers msg))
      #t
      #f))

;;; Utility for writing to a port

(define-method (to-port (msg <http-message>) sock)
  (write-prologue msg sock)
  (write-headers msg sock)
  (write-separator sock)
  (write-content msg sock))

;; XXX: change to use regexes instead of string indices.

(define-method (read-headers (msg <http-message>) sock)
  (let read-header ((line (read-crlf-line sock)))
    (if (string-null? line)
	line
	(let* ((colon-idx (string-index line #\072))
	       (name (substring line 0 colon-idx))
	       (space-idx (string-index line #\space colon-idx))
	       (value (substring line (+ 1 space-idx))))
	  (header-set! msg name value)
	  (read-header (read-crlf-line sock))))))

;;; Write out the separator (\r\n) to the receiving port

(define (write-separator sock)
  (format sock "\r\n"))

;;; Copy data from content to the receiving port

(define-method (write-content (msg <http-message>) sock)
  (let ((content (content msg)))
    (if (null? content)
	'()
        (begin
          (copy-port content sock)
          (close content)))))

;;; Write the headers.

(define-method (write-headers (msg <http-message>) sock)
  (let write-all-headers ((headers (headers msg)))
    (if (null? headers)
	'()
	(begin
	  (format sock "~A: ~A\r\n" (car (car headers)) (cdr (car headers)))
	  (write-all-headers (cdr headers))))))

;;; Class defining an HTTP request.

(define-class <http-request> (<http-message>)
  (uri #:accessor uri			; string
       #:init-keyword #:uri)
  (server #:accessor server		; string, hostname of server to call
	  #:init-keyword #:server)
  (port #:accessor port                 ; int, port to use
	#:init-keyword #:port)
  (method #:init-value "GET"		; string
          #:accessor meth               ; Try changing this to method, for fun.
          #:init-keyword #:method))

;;; Send the request, and get back a response.

(define-method (send-request (req <http-request>))
  (let ((sock (connect-request req)))
    (to-port req sock)
    (let ((resp (response-from-port sock)))
      resp)))

(define-method (request-from-port sock)
  (let ((req (make <http-request>)))
    (read-command req sock)
    (read-headers req sock)
    (set! (content req) sock)
    req))


;;; Connect to the appropriate host and port for the given request.
;;; Returns the socket connection.

(define-method (connect-request (req <http-request>))
  (let* ((host (gethostbyname (server req)))
	 (addr (car (hostent:addr-list host)))
	 (sock (socket AF_INET SOCK_STREAM 0)))
    (connect sock AF_INET addr (port req))
    sock))

;;; Write the command part of the request to the receiving port.

(define-method (write-prologue (req <http-request>) sock)
  (write-command req sock))

(define-method (write-command (req <http-request>) sock)
  (format sock "~A ~A HTTP/~A\r\n" (meth req) (uri req) (version req)))

(define-method (read-command (req <http-request>) sock)
  (let* ((command-line (read-crlf-line sock))
         (match (string-match "([a-zA-Z]+) ([^ ]+) HTTP/([0-9.]+)" command-line)))
    (if (null? match)
        #f
        (begin
          (set! (meth req) (match:substring match 1))
          (set! (uri req) (match:substring match 2))
          (set! (version req) (match:substring match 3))
          #t))))

;;; Class defining an HTTP response

(define-class <http-response> (<http-message>)
  (code #:accessor code
	#:init-keyword #:code)
  (status #:accessor status
	  #:init-keyword #:status))

(define (response-from-port sock)
  (let ((resp (make <http-response>)))
    (read-status resp sock)
    (read-headers resp sock)
    (set! (content resp) sock)
    resp))

(define-method (write-prologue (resp <http-response>) sock)
  (write-status resp sock))

(define-method (write-status (resp <http-response>) sock)
  (format sock "HTTP/~A ~A ~A\r\n" (version resp) (code resp) (status resp)))

(define-method (read-status (resp <http-response>) sock)
  (let* ((status-line (read-crlf-line sock))
	 (vidx (string-index status-line #\space)))
    (if (not vidx)
	(throw 'bad-status-line status-line)
	(let* ((idx (string-index status-line #\space (+ 1 vidx)))
	       (codestr (substring status-line (+ 1 vidx) idx))
	       (code-value (string->number codestr))
	       (status-value (substring status-line (+ 1 idx))))
	  (set! (code resp) code-value)
	  (set! (status resp) status-value)
	  status-line))))

(define (read-crlf-line port)
  (let* ((line (read-line port))
	 (cridx (string-index line #\cr)))
    (if cridx
	(substring line 0 cridx)
	line)))
